/*
 *  Copyright (C) 2000  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include "mac_if.h"
#include "dos.h"
#include "memory.h"

unsigned    nmbfre;
unsigned    nmblks;
char        Xpar[L2PNUM];
char        Dpar[L2PNUM];
char        Tpar[L2PNUM];
char        Ppar[L2PNUM];
char        Wpar[L2PNUM];
char        Cpar[L2PNUM];
unsigned    TApar[L2PNUM];
admin_t     lnktbl[L2PNUM];
char        ishmod;
unsigned    ticks;

pool_t *next_free;              /* Next free buffer in the buffer pool      */
pool_t *rx_buf;

pool_t *allocb()
{
    pool_t *bp = NULL;

    /* Allocate buffer */
    idisable();
    if(nmbfre != 0)
    {
        bp = next_free;
        next_free = bp->next;
        nmbfre--;
    }
    irestore();

    return bp;
}

void dealoc(pool_t *bp)
{
    /* Free buffer */
    idisable();
    bp->next  = next_free;
    next_free = bp;
    nmbfre++;
    irestore();
}

/**************************************************************************\
*                                                                          *
* initialize parameters                                                    *
*                                                                          *
\**************************************************************************/
void init()
{
    unsigned  int n;
    char     *actbp;
    char     *nextbp;
    char     *minm;         /* Lowest available address  */
    char     *maxm;         /* Highest available address */
    admin_t  *linkp;

    for (n = 0; n < L2PNUM; n++)
    {
        Tpar[n]  = defTp;
        Wpar[n]  = defWp;
        Ppar[n]  = defPp;
        Cpar[n]  = defCp;
        Dpar[n]  = defDp;
        TApar[n] = defTAp;
        Xpar[n]  = defXp;

        linkp->tx_buf = NULL;
        linkp->tx_act = NULL;
        linkp->rx_act = NULL;
        linkp->tx_pos = 0;
        linkp->rx_pos = 0;
    }
    rx_buf = NULL;

    ishmod = 0;
    ticks = 0;

    nmblks = nmbfre = 0;      /* keine Links aktiv, noch keine Freibuffer */
    minm = actbp = minmem();  /* niedrigste nutzbare Freispeicherstelle   */
    maxm = maxmem();          /* hoechste nutzbare Freispeicherstelle     */

    next_free = NULL;
    while ( ( (nextbp = actbp + sizeof(pool_t) - 1) <= maxm )
            &&
            ( nextbp >= minm )
          )                                  /* solange Buffer in Frei-   */
    {                                        /* speicher passen..         */
        dealoc(actbp);                       /* initialisieren und in     */
        actbp = nextbp + 1;                  /* Freispeicherliste haengen */
    }
}

/**************************************************************************\
*                                                                          *
* "main function"                                                          *
*                                                                          *
\**************************************************************************/
void mainf()
{
    init();
}

void l2rest(void)
{
}

int /* boolean */ info_avl(void)
{
}

void lx(void)
{
}

void l2rx(void)
{
}

void l2tx(void)
{
    char port;

    /* kick all the ports which have data pending */
    for(port = 0; port < l1ports; port++)
    {
        if(lnktbl[port].tx_buf != NULL)
        {
            kicktx(port);
        }
    }
}

/* used by ax25_mac.c to pass received characters */
unsigned int l1put(unsigned int action)
{
    unsigned int  port;
    unsigned int  res;
    admin_t      *linkp;
    pool_t       *p;

    port = (action >> 8) & 0x007F;
    if(port >= L2PNUM)
    {
        return 1; /* skip this fame */
    }

    linkp = &lnktbl[port];

    if((action & 0x8000) == 0)
    {
        if(linkp->rx_act == NULL)
        {
            p = allocb();
            if(p != NULL)
            {
                p->next = NULL;
                linkp->rx_act = p;
                linkp->rx_pos = 0;
                linkp->rx_act->l1frame.port = port;
            }
            else
            {
                return 1; /*  skip this frame */
            }
        }
        linkp->rx_act->l1frame.frame[linkp->rx_pos] = (action & 0x00ff);
        linkp->rx_pos++;
        if((linkp->rx_pos) >= MAXFLEN) /*  check overflow */
        {
            /* discard this frane */
            dealoc(linkp->rx_act);
            linkp->rx_act = NULL;
            linkp->rx_pos = 0;
            return 1; /*  skip this frame */
        }
    }
    else
    {
        if(linkp->rx_act != NULL)
        {
            if((action & 0x0001) != 0)
            {
                /* discard this frane */
                dealoc(linkp->rx_act);
                linkp->rx_act = NULL;
                linkp->rx_pos = 0;
                return 0; /* no problem */
            }
            else
            {
                /*  frane complete, add to rx_buf chain */
                if(linkp->rx_pos > 2)
                {
                    /*  truncate frame-check garbage */
                    linkp->rx_pos -= (action >> 1) & 0x7F;
                }
                linkp->rx_act->l1frame.len = linkp->rx_pos;

                idisable();
                p = linkp->rx_act;
                p->next = rx_buf;
                rx_buf = p;
                irestore();

                linkp->rx_act = NULL;
                linkp->rx_pos = 0;
            }
        }
    }
    return 0; /* no problem */
}

/* used by ax25_mac.c to retrieve characters to send */
unsigned int l1get(int action)
{
    unsigned int  port;
    unsigned int  res;
    admin_t      *linkp;
    pool_t       *p,*q;

    port = (action >> 8) & 0x007F;
    if(port >= L2PNUM)
    {
        return 0x8001; /* no more data */
    }

    linkp = &lnktbl[port];

    if((action & 0x8000) == 0)
    {
        if(linkp->tx_act == NULL)
        {
            idisable();
            p = linkp->tx_buf;
            if(p != NULL)
            {
                linkp->tx_buf = p->next;
                p->next = NULL;
            }
            irestore();
            if(p != NULL)
            {
                linkp->tx_act = p;
                linkp->tx_pos = 0;
            }
            else
            {
                return 0x8001; /* no more data */
            }
        }

        if(linkp->tx_pos < linkp->tx_act->l1frame.len)
        {
            res = linkp->tx_act->l1frame.frame[linkp->tx_pos];
            linkp->tx_pos++;
            return res;
        }
        else
        {
            /* discard this frane, all send */
            dealoc(linkp->tx_act);
            linkp->tx_act = NULL;
            linkp->tx_pos = 0;

            if(linkp->tx_buf != NULL)
            {
                return 0x8000; /* end of frame, next frame waiting */
            }
            else
            {
                return 0x8001; /* no more data */
            }
        }
    }
    else
    {
        if((action & 0x0001) == 0)
        {
            /* rewind */
            if(linkp->tx_act != NULL)
            {
                linkp->tx_pos = 0; /* rewind to start of frame */
            }
        }
        else
        {
            /* discard */
            if(linkp->tx_act != NULL)
            {
                /* discard this frane */
                dealoc(linkp->tx_act);
                linkp->tx_act = NULL;
                linkp->tx_pos = 0;
            }

            /* discard all pending tx-frames */
            idisable();
            p = linkp->tx_buf;
            linkp->tx_buf = NULL;
            irestore();

            while(p != NULL)
            {
                q = p->next;
                dealoc(p);
                p = q;
            }
        }
    }
}

void __cdecl __interrupt __far ax25_state (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    int flag = 0;

    idisable();
    if(clock_error == true) set_pit();
    if(rx_buf != NULL)
    {
        flag = 1;
    }
    irestore();

    _ax = flag;
}

void __cdecl __interrupt __far ax25_input (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    pool_t * buf, * pbuf;
    frame_t far * fp;

	fp = MK_FP(_es, _di);

	idisable();

	buf = rx_buf;
	if(buf != NULL)
	{
		pbuf = NULL;
		while(buf->next != NULL)
		{
			pbuf = buf;
			buf = pbuf->next;
		}
		if(pbuf == NULL)
		{
			rx_buf = NULL; /*  no more frames in rx_buf */
		}
		else
		{
			pbuf->next = NULL; /* pbuf is the last buffer in rx_buf */
		}
	}

	irestore();

	if(buf == NULL)
	{
		_ax = 0x00FF;
		return;
	}

	fp->port = buf->l1frame.port;
	fp->len  = buf->l1frame.len;
	_fmemcpy((char far *) &(fp->frame[0]),
             (char far *) &(buf->l1frame.frame[0]), buf->l1frame.len);

	dealoc(buf);

	_ax = 0;
}

void __cdecl __interrupt __far ax25_output (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    pool_t      * buf, * pbuf;
    frame_t far * fp;
    admin_t     *linkp;
    char        port;

	fp = MK_FP(_es, _di);

    buf = allocb();

    if(buf == NULL)
    {
        /* out of buffers... */
		_ax = 0x00FF;
		return;
    }

    /* copy the information to this new buffer */
    port = fp->port;
    buf->l1frame.port = port;
    buf->l1frame.len  = fp->len;
    _fmemcpy((char far *) &(buf->l1frame.frame[0]),
             (char far *) &(fp->frame[0]), fp->len);
    buf->next = NULL;

    linkp = &lnktbl[port];

    /* now add the buffer to the end of the tx_buf list */
    idisable();

    pbuf = linkp->tx_buf;
    if(pbuf != NULL)
	{
        while(pbuf->next != NULL)
		{
            pbuf = pbuf->next;
		}
        pbuf->next = buf; /* add buf at the end of the chain */
	}
    else
    {
        linkp->tx_buf = buf; /*  first frame in tx_buf */
    }

    irestore();

    kicktx(port);

	_ax = 0;
}

void __cdecl __interrupt __far ax25_debug (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    pool_t     *bp;

    _dx = 0;
    _bx = 0;

    /* count free buffers */
    idisable();
    bp = next_free;
    while(bp != NULL)
    {
        if(bp < minmem)
        {
            _bx = 1;
            break;
        }
        else if(bp < minmem)
        {
            _bx = 2;
            break;
        }

        bp = bp->next;
        _dx++;
        if(_dx == 0xffff)
        {
            _bx = 3;
            break;
        }
    }
    _cx = nmbfre; /* what the SW thinks */
    irestore();
    
	_ax = 0;
}

void __cdecl __interrupt __far ax25_bufdbg (
  unsigned _es, unsigned _ds,
  unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp,
  unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax,
  unsigned _ip, unsigned _cs,
  unsigned flags )
{
    debug_t      dbg;
    debug_t far *fp;
    int          i;
    admin_t     *linkp;

	fp = MK_FP(_es, _di);

    dbg.buf_free  = nmbfre;
    dbg.next_free = (char far *) &next_free;
    dbg.rx_buf    = (char far *) &(rx_buf);
    dbg.minmem    = (char far *) minmem();
    dbg.maxmem    = (char far *) maxmem();
    for(i = 0; i < 8; i++)
    {
        linkp = &lnktbl[i];

        dbg.rx_act[i] = (char far *) &(linkp->rx_act);
        dbg.rx_pos[i] = (char far *) &(linkp->rx_pos);
        dbg.tx_buf[i] = (char far *) &(linkp->tx_buf);
        dbg.tx_act[i] = (char far *) &(linkp->tx_act);
        dbg.tx_pos[i] = (char far *) &(linkp->tx_pos);
    }

	_fmemcpy((char far *) fp, (char far *) &dbg, sizeof(debug_t));

	_ax = 0;
}
